package com.r3.corda.networkmanage.hsm

import com.r3.corda.networkmanage.common.persistence.configureDatabase
import com.r3.corda.networkmanage.common.utils.ShowHelpException
import com.r3.corda.networkmanage.common.utils.initialiseSerialization
import com.r3.corda.networkmanage.hsm.configuration.parseParameters
import com.r3.corda.networkmanage.hsm.processor.CsrProcessor
import com.r3.corda.networkmanage.hsm.processor.NetworkMapProcessor
import org.apache.logging.log4j.LogManager
import org.bouncycastle.jce.provider.BouncyCastleProvider
import java.security.Security
import javax.crypto.Cipher

private val logger = LogManager.getLogger("com.r3.corda.networkmanage.hsm.Main")

fun main(args: Array<String>) {
    parseParameters(*args).run {
        try {
            // Validate
            // Grabbed from https://stackoverflow.com/questions/7953567/checking-if-unlimited-cryptography-is-available
            require(Cipher.getMaxAllowedKeyLength("AES") >= 256) {
                "Unlimited Strength Jurisdiction Policy Files must be installed, see http://www.oracle.com/technetwork/java/javase/downloads/jce8-download-2133166.html"
            }
            require(csrSigning != null || networkMapSigning != null) {
                "Either network map or certificate signing request certificate parameters must be specified."
            }
            requireNotNull(dataSourceProperties)

            // Ensure the BouncyCastle provider is installed
            if (Security.getProvider(BouncyCastleProvider.PROVIDER_NAME) == null) {
                Security.addProvider(BouncyCastleProvider())
            }

            initialiseSerialization()
            // Create DB connection.
            val persistence = configureDatabase(dataSourceProperties, database)
            if (networkMapSigning != null) {
                NetworkMapProcessor(networkMapSigning, device, keySpecifier, persistence).run()
            }
            if (csrSigning != null) {
                try {
                    CsrProcessor(csrSigning, device, keySpecifier, persistence).showMenu()
                } catch (e: ShowHelpException) {
                    e.errorMessage?.let(::println)
                    e.parser.printHelpOn(System.out)
                }
            }
        } catch (e: Exception) {
            logger.error("Error while starting the HSM Signing service.", e)
        }
    }
}


